"""
seafog.himawari provides methods to download Himawari-8/9 data from PTREE and to plot true color satellite image.
"""

from datetime import datetime
from os import makedirs
from os.path import exists
from typing import Callable, Tuple, Union

import numpy as np
from cartopy import crs
from cartopy.mpl.geoaxes import GeoAxes
from numpy import ndarray
from rich.progress import Progress
from scipy.interpolate import interp1d
from scipy.ndimage import zoom
from xarray import Dataset

from .utils import download_ftp, logger


def convert_satellite_data(satellite_data: ndarray, min_value: float = 0, max_value: float = 1) -> ndarray:
    """
    transform satellite data from albedo value to RGB value (0-255) with segmentation mapping.
    
    :param satellite_data: satellite data
    :param min_value: the minimum albedo value that will be mapped to 0
    :param max_value: the maximum albedo value that will be mapped to 255
    :return: RGB values.

    >>> data = np.linspace(0, 1, 10)
    >>> convert_satellite_data(data)
    array([  0, 103, 154, 180, 204, 219, 231, 241, 248, 255])

    """
    satellite_data = (satellite_data - min_value) / (max_value - min_value) * 255
    # satellite_data = satellite_data.astype(int)

    # segmentation mapping
    x = [0, 30, 60, 120, 190, 255]
    y = [0, 110, 160, 210, 240, 255]
    interpolator = interp1d(x, y, bounds_error=False, fill_value=255)

    return interpolator(satellite_data).astype(int)


def up_sampling(data: ndarray, target_size: Tuple[int, int]) -> ndarray:
    """
    up sampling a image.
    
    :param data: image data.
    :param target_size: target shape.
    :return: new image data.
    """
    zoom_scale = []
    for raw_len, target_len in zip(data.shape, target_size):
        zoom_scale.append(target_len / raw_len)

    return zoom(data, zoom_scale, order=3)


def plot_himawari_vis(ax: GeoAxes, dataset: Dataset, area_range: Union[Tuple[float, ...], Tuple[int, ...], None] = None, *args, **kwargs):
    """
    plot VIS image with giving satellite data.
    
    :param ax: cartopy axes object.
    :param dataset: himawari data.
    :param area_range: area range, (lon_min, lon_max, lat_min, lat_max).
                       if None, plot the whole area.
    :param args: other params that will be passed to ``seafog.convert_satellite_data``.
    :param kwargs: other params that will be passed to ``seafog.convert_satellite_data``.
    :return:
    """
    # check axes
    assert isinstance(ax, GeoAxes), "`ax` must be cartopy.mpl.geoaxes.GeoAxes so we can draw VIS image."

    # check data bands
    for band in ["albedo_01", "albedo_02", "albedo_03"]:
        assert band in dataset.data_vars, f"We need `{band}` band to draw VIS image."

    # check area_range
    if area_range is not None:
        # extract area range
        lon_min, lon_max, lat_min, lat_max = area_range
    else:
        lon_min, lon_max, lat_min, lat_max = None, None, None, None

    # area index
    index = (slice(lat_max, lat_min), slice(lon_min, lon_max))

    # extract RGB bands
    blue_band = dataset["albedo_01"].loc[index]
    green_band = dataset["albedo_02"].loc[index]
    red_band = dataset["albedo_03"].loc[index]

    # extract longitude and latitude
    longitude = blue_band["longitude"]
    latitude = blue_band["latitude"]

    # create an RGB image array
    blue_value = convert_satellite_data(blue_band.to_numpy(), *args, **kwargs)
    green_value = convert_satellite_data(green_band.to_numpy(), *args, **kwargs)
    red_value = convert_satellite_data(red_band.to_numpy(), *args, **kwargs)
    vis_rgb = np.dstack((red_value, green_value, blue_value)) / 255

    # create a color map
    vis_color = vis_rgb.reshape(-1, 3)

    # plot
    ax.pcolormesh(longitude, latitude, vis_rgb, color=vis_color, shading="nearest", transform=crs.PlateCarree())


def himawari_find_data(date: str, save_path: str, resolution: str = "low", area: str = "full", band_num: str = "auto", ftp_setting: dict = None, progress_num=1,
                       progress: Progress = None, show_progress=True, callback: Callable = None, timeout=720, connection_timeout=30, force_download=False) -> str:
    """
    Download himawari-8/9 data from the official website.
    This function will search the data in the ``save_path`` first.
    If the data exists, it will skip the download.
    You can set `force_download=True` to force to re-download the data.

    :param date: Date string in format "%Y-%m-%d %H:%M".
    :type date: str
    :param save_path: The directory to store data.
    :type save_path: str
    :param resolution: Himawari data's resolution, valid value: ``["low", "high"]``.
                       The actual resolution will be different depending on the ``area`` you chose.
                       You can refer to the `seafog wiki <http://gitea.seafog.syize.cn/seafog/seafog-plot/wiki/himawari8#himawari8_find_data>`_ for more information.
    :type resolution: str
    :param area: Himawari data's area, valid value: ``["full", "japan"]``.
    :type area: str
    :param band_num: Band numbers in the data, valid value: ``["21", "04", "14", "auto"]``.
                     The band number will change with the ``area``, ``full`` is ``"21"``, ``japan`` is ``"04"`` or ``"14"``.
                     If ``band_num == "auto"``, the usable highest number will be used.
    :type band_num: str
    :param ftp_setting: A dict contains the following keys: ``["username", "password", "proxy_host", "proxy_port"]``, in which ``["proxy_host", "proxy_port"]`` are optional.
    :type ftp_setting: dict
    :param progress_num: The number of processes which will be used to download the data.
                         Increase this number can help make download speed faster.
                         It won't be larger than ``cpu_cores – 2``.
    :type progress_num: int
    :param progress: ``rich.progress.Progress`` object to display progress bar.
    :type progress: Progress
    :param show_progress: If True, display download progress in the console.
    :type show_progress: bool
    :param callback: Callable object that will be called every iteration during data download.
                     ``callback`` should accept two params, `total_size` and `step_size`.
    :type callback: Callable
    :param timeout: The max time which could be used to download the data.
                    Download will be terminated once the max time is reached.
                    Set it to -1 for no limitation.
                    Units: seconds.
    :type timeout: int
    :param connection_timeout: The max time which will be used to connect to the server.
                               Units: seconds.
    :type connection_timeout: int
    :param force_download: If ignore the existing data and re-download it.
    :type force_download: bool
    :return: Data file path.
    :rtype: str
    """
    # for example, NC_H08_20150707_0000_r04_FLDK.05401_05201.nc
    NAME_TEMPLATE = "NC_H{}_{}_{}_{}_FLDK.{}_{}.nc"
    # for example: ftp://ftp.ptree.jaxa.jp/jma/netcdf/201507/07/NC_H08_20150707_0010_R21_FLDK.06001_06001.nc
    URL_TEMPLATE = "ftp://ftp.ptree.jaxa.jp/jma/netcdf/{}/{}/"

    resolution_dict = {
        'full': {
            'low': {'x': '02401', 'y': '02401'},
            'high': {'x': '06001', 'y': '06001'}
        },
        'japan': {
            'low': {'x': '02701', 'y': '02601'},
            'high': {'x': '05401', 'y': '05201'}
        }
    }

    # check date
    if datetime.strptime(date, "%Y-%m-%d %H:%M") < datetime(2015, 7, 7, 0, 0):
        logger.error(f"Date is earlier than 2015-07-07 00:00")
        raise ValueError

    # check area
    if area not in ['full', 'japan']:
        logger.error(f"Unknown area: {area}")
        raise ValueError

    # check resolution
    if resolution not in ['low', 'high']:
        logger.error(f"Unknown resolution: {resolution}")
        raise ValueError

    # check save path
    if not exists(save_path):
        makedirs(save_path)

    # check band number
    if band_num == 'auto':
        if area == 'full':
            band_num = "R21"
        else:
            band_num = "r14"
    else:
        if area == "full":
            band_num = "R21"
        else:
            if band_num not in ["04", "14"]:
                logger.error(f"Error band number: should be '04' or '14', but is {band_num}")
                raise ValueError
            band_num = f"r{band_num}"

    x_res = resolution_dict[area][resolution]['x']
    y_res = resolution_dict[area][resolution]['y']

    # pares date
    date = datetime.strptime(date, "%Y-%m-%d %H:%M")

    # generate url and save path
    satellite_num = "08" if date <= datetime(2022, 12, 12) else "09"
    filename = NAME_TEMPLATE.format(satellite_num, date.strftime("%Y%m%d"), date.strftime("%H%M"), band_num, x_res, y_res)
    data_path = f"{save_path}/{filename}"
    # check the existing data
    if not force_download and exists(data_path):
        return data_path

    url = URL_TEMPLATE.format(date.strftime("%Y%m"), date.strftime("%d")) + filename

    # parse ftp setting
    user = None
    passwd = None
    proxy_host = None
    proxy_port = None

    if ftp_setting is not None:
        for key in ftp_setting:
            if key == "username":
                user = ftp_setting[key]
            elif key == "password":
                passwd = ftp_setting[key]
            elif key == "proxy_host":
                proxy_host = ftp_setting[key]
            elif key == "proxy_port":
                proxy_port = ftp_setting[key]

    # inform user to apply P-Tree account
    if None in [user, passwd]:
        logger.error(f"You doesn't give your username and password of PTREE, which is necessary to download satellite data. You can apply an account in this website: "
                     f"https://www.eorc.jaxa.jp/ptree/registration_top.html")
        raise KeyError

    if not download_ftp(url, save_path, filename, user=user, passwd=passwd, proxy_host=proxy_host, proxy_port=proxy_port,
                        show_progress=show_progress, progress=progress, callback=callback, timeout=timeout, connection_timeout=connection_timeout,
                        progress_num=progress_num, force_download=force_download):
        raise ConnectionError

    return f"{save_path}/{filename}"


__all__ = ['convert_satellite_data', 'up_sampling', 'himawari_find_data', 'plot_himawari_vis']



"""
seafog.detect contains Seafog class which provides several methods to detect seafog top height and seafog area.
"""

from time import ctime
from typing import Union, Tuple, List

import numpy as np
from xarray import DataArray, Dataset

from .utils import solar_altitude_zenith, logger, seafog_top_height


def remove_single_point(data: np.ndarray) -> np.ndarray:
    """
    remove single point which eight surrounding points' valid number is less than 4 or the 16 surround-surrounding points' valid number is less than 8.

    :param data: 2-D ndarray.
    :return: data with single points has been removed.
    """
    result = data.copy()

    # loop and check
    for i in range(2, data.shape[0] - 2):
        for j in range(2, data.shape[1] - 2):
            # if the point is np.nan, pass
            if np.isnan(data[i, j]):
                continue
            # check the innermost ring's 8 points and outermost ring's 16 points
            index_8 = np.where(~np.isnan(data[i - 1:i + 2, j - 1:j + 2]))
            index_16 = np.where(~np.isnan(data[i - 2:i + 3, j - 2:j + 3]))
            if index_8[0].size < 5:
                result[i, j] = np.nan
            elif index_16[0].size - 9 < 8:
                result[i, j] = np.nan

    return result


class Seafog:
    """
    ``Seafog`` class provides several methods to detect seafog top height and seafog area.
    seafog.detect module has exported a Seafog instance called ``detect_seafog``, use it directly.
    """

    def __init__(self):
        # place parameters setting here
        self._diff_range_night_IR: Tuple[float, float] = (-5.5, -3.5)
        self._diff_range_daytime_IR: Tuple[float, float, float] = (-2, 3, 45)
        self._beta: float = 0.0685
        self._ir_sst_diff_range: Tuple[float, float] = (-5, 10)
        # threshold value of QCLOUD
        self._qcloud_threshold = 0.02
        # datas which are higher than 500m will be removed
        self._maximum_height = 600

    # ############################################## parameters setting ##############################################

    @property
    def diff_range_night_IR(self) -> Tuple[float, float]:
        """
        the value range of ``IR4 - IR1`` at night.

        :return: (min_value, max_value)
        """
        return self._diff_range_night_IR

    @diff_range_night_IR.setter
    def diff_range_night_IR(self, value: Union[Tuple[float, float], List[float]]):
        assert len(value) == 2, f"Only two values will be accepted, but got {value}"
        assert isinstance(value[0], float) and isinstance(value[1], float), f"Only float value will be accepted, but got {value}"

        self._diff_range_night_IR = tuple(value)

    @property
    def diff_range_daytime_IR(self) -> Tuple[float, float, float]:
        """
        the value range of ``IR4 - IR1`` in the daytime, including three values of two range.

        :return: (start, end(start_two), end_two)
        """
        return self._diff_range_daytime_IR

    @diff_range_daytime_IR.setter
    def diff_range_daytime_IR(self, value: Union[Tuple[float, float, float], List[float]]):
        assert len(value) == 3, f"Only three values will be accepted, but got{value}"
        for _value in value:
            assert isinstance(_value, float) or isinstance(_value, int), f"Only int or float value will be accepted, but got {value}"

        value = tuple(float(x) for x in value)
        self._diff_range_daytime_IR = value

    @property
    def beta(self) -> float:
        """
        back scatter parameter.

        :return: back scatter parameter.
        """
        return self._beta

    @beta.setter
    def beta(self, value: float):
        assert isinstance(value, float), "Only float value will be accepted"
        self._beta = value

    @property
    def IR_sst_diff_range(self) -> Tuple[float, float]:
        """
        the value range of ``IR4 - SST``.

        :return: (min_value, max_value).
        """
        return self._ir_sst_diff_range

    @IR_sst_diff_range.setter
    def IR_sst_diff_range(self, value: Tuple[float, float]):
        assert isinstance(value, tuple), "Only float value will be accepted"
        assert len(value) == 2, f"Only two values will be accepted, but got{value}"
        self._ir_sst_diff_range = value

    @property
    def QCLOUD_threshold(self) -> float:
        """
        liquid water content of the cloud.
        any part of air which QCLOUD is bigger than this threshold will be regarded as seafog.

        :return: threshold value of QCLOUD.
        """
        return self._qcloud_threshold

    @QCLOUD_threshold.setter
    def QCLOUD_threshold(self, value: float):
        self._qcloud_threshold = value

    @property
    def maximum_height(self) -> int:
        """
        maximum height of seafog top.

        :return: threshold value.
        """
        return self._maximum_height

    @maximum_height.setter
    def maximum_height(self, value: int):
        self._maximum_height = value

    # ################################################################################################################

    def seafog_from_wrfout(self, wrfout: Dataset) -> Dataset:
        """
        calculate the seafog area and top height of seafog using WRF data.

        :param wrfout: WRF data
        :return: ``xarray.Dataset`` object.
        """
        # check data in dataset
        for field in ["QCLOUD", "XTIME", "HGT", "PH", "PHB", "XLONG", "XLAT"]:
            assert field in wrfout.keys(), f"Your WRF data doesn't have {field} data, which is necessary for calculation"

        # read QCLOUD, change unit to g/kg
        QCloud = wrfout["QCLOUD"].to_numpy() * 1000
        # read other data
        terrain_height = wrfout["HGT"].to_numpy()
        perturbation_geo = wrfout["PH"].to_numpy()
        base_geo = wrfout["PHB"].to_numpy()

        # expand terrain_height's dimension
        terrain_height = np.expand_dims(terrain_height, axis=1)
        terrain_height = np.tile(terrain_height, (1, perturbation_geo.shape[1], 1, 1))

        # calculate height from sea level
        height = (perturbation_geo + base_geo) / 9.8 - terrain_height
        # calculate half-height
        height = (height[:, 1:, :, :] + height[:, :-1, :, :]) / 2

        # search datas which are under maximum height and bigger that QCLOUD threshold
        height_index = height <= self._maximum_height
        QCloud_index = QCloud >= self._qcloud_threshold
        # combine and get seafog index
        seafog_index = np.logical_and(height_index, QCloud_index)

        # then we need to calculate the height where QCLOUD == QCLOUD_threshold, using linear interpolation
        # # copy height
        seafog = height.copy()
        # # change datas which index is False to nan
        seafog[~seafog_index] = np.nan
        # # find the largest height values
        valid_largest_height = np.nanmax(seafog, axis=1)
        # # expand dimension
        valid_largest_height = np.expand_dims(valid_largest_height, axis=1)
        # # get indexes of these values
        valid_index = seafog == valid_largest_height

        # create a new height array and new QCloud array to get values at a higher level
        array_shape = list(height.shape)
        array_shape[1] = 1
        fake_array = np.zeros(array_shape)
        new_height = np.concatenate([height[:, 1:, :, :], fake_array], axis=1)
        new_QCloud = np.concatenate([QCloud[:, 1:, :, :], fake_array], axis=1)

        # take out values with these two indexes
        valid_height_value = height[valid_index]
        valid_QCloud_value = QCloud[valid_index]
        invalid_height_value = new_height[valid_index]
        invalid_QCloud_value = new_QCloud[valid_index]

        # interpolate
        QCloud_gradient = (invalid_height_value - valid_height_value) / (invalid_QCloud_value - valid_QCloud_value)
        delta_height = (self._qcloud_threshold - valid_QCloud_value) * QCloud_gradient
        # if delta_height is larger than 50, change it to 50
        delta_height[delta_height > 50] = 50.

        # put it to the original height
        height[valid_index] = height[valid_index] + delta_height

        # change datas which index is False to nan
        height[~seafog_index] = np.nan

        # find the largest height value, which is seafog's top
        height = np.nanmax(height, axis=1)

        # get seafog's bottom
        QCloud_surface = QCloud[:, 0, :, :]

        # create DataArray and return
        longitude = wrfout["XLONG"].to_numpy()[0]
        latitude = wrfout["XLAT"].to_numpy()[0]
        time = wrfout["XTIME"].to_numpy()

        longitude_index = np.arange(longitude.shape[1])
        latitude_index = np.arange(longitude.shape[0])

        create_time = ctime()
        return Dataset(data_vars={
            "seafog": DataArray(name="seafog", data=height, dims=["time", "latitude_index", "longitude_index"], coords={
                "time": time,
                "latitude_index": latitude_index,
                "longitude_index": longitude_index
            }, attrs={
                "units": "m",
                "long_name": "height of seafog's top from WRF",
                "standard_name": "seafog",
                "description": f"height of seafog's top from WRF, calculated by package seafog at {create_time}"
            }),
            "seafog_surface": DataArray(name="seafog_surface", data=QCloud_surface, dims=["time", "latitude_index", "longitude_index"], coords={
                "time": time,
                "latitude_index": latitude_index,
                "longitude_index": longitude_index
            }, attrs={
                "units": "g/kg",
                "long_name": "QCLOUD at surface from WRF",
                "standard_name": "seafog_surface"
            }),

            "longitude": DataArray(name="longitude", data=longitude, dims=["latitude_index", "longitude_index"], coords={
                "latitude_index": latitude_index,
                "longitude_index": longitude_index
            }, attrs={
                "units": "degree",
                "long_name": "longitude",
                "standard_name": "longitude"
            }),
            "latitude": DataArray(name="latitude", data=latitude, dims=["latitude_index", "longitude_index"], coords={
                "latitude_index": latitude_index,
                "longitude_index": longitude_index
            }, attrs={
                "units": "degree",
                "long_name": "latitude",
                "standard_name": "latitude"
            })
        }, attrs={
            "description": f"height of seafog's top and area of seafog bottom from WRF, calculated by package seafog at {create_time}"
        })

    def seafog_top_daily(self, VIS: np.ndarray, zenith_cosine: np.ndarray) -> np.ndarray:
        """
        calculate the top height of seafog in the daytime.

        :param VIS: VIS band data.
        :param zenith_cosine: cosine value of the solar zenith angle.
        :return:
        """
        return -45.6 + 84.3 * np.power(VIS / 100 * zenith_cosine / ((1 - VIS / 100) * self._beta), 0.5)

    def daytime_seafog(self, date: str, IR4: DataArray, VIS: DataArray, sst: DataArray) -> DataArray:
        """
        Detect seafog based on the difference between IR4 and sst data.

        :param date: Date of the data in format ``%Y-%m-%d %H:%M``, UTC.
        :param IR4: IR4 band data.
        :param VIS: Brightness temperature of the VIS band.
        :param sst: SST data.
        :return: Height of the marine fog top.
        """
        # check data dimensions
        assert (IR4.shape[0] == VIS.shape[0] == sst.shape[0]), "The shape of IR4 and SST data must be the same"
        assert (IR4.shape[1] == VIS.shape[1] == sst.shape[1]), "The shape of IR4 and SST data must be the same"

        # detect fog
        IR_diff = IR4.to_numpy() - sst.to_numpy() - 4.0
        index = np.logical_and(IR_diff >= self._ir_sst_diff_range[0], IR_diff <= self._ir_sst_diff_range[1])
        IR_diff[~index] = np.nan
        index = np.isnan(IR_diff)

        # calculate fog top height
        fog_top_height = VIS.to_numpy()
        fog_top_height[index] = np.nan
        longitude = VIS['longitude'].to_numpy()
        latitude = VIS['latitude'].to_numpy()
        longitude, latitude = np.meshgrid(longitude, latitude)
        _, solar_zenith = solar_altitude_zenith(date, longitude, latitude)
        fog_top_height = seafog_top_height(fog_top_height, solar_zenith, self._beta)

        return DataArray(name="seafog", data=fog_top_height, dims=VIS.dims, coords=VIS.coords, attrs={
            "units": "m",
            'long_name': "height of seafog's top",
            'standard_name': 'seafog',
            'description': f"height of seafog's top, calculated by package seafog at {ctime()}"
        })

    def night_seafog(self, date: str, IR1: DataArray, IR4: DataArray) -> DataArray:
        """
        detect seafog area and thickness based on the IR1, IR4 and VIS satellite data and GOOS sst data.

        :param date: UTC date of the data, for example, '2020-05-19 09:13'.
        :param IR1: IR1 data.
        :param IR4: IR4 data.
        :return: detected seafog height.
        """
        # check data dimensions
        assert (IR1.shape[0] == IR4.shape[0]), "The shape of IR1 and IR4 data must be the same"
        assert (IR1.shape[1] == IR4.shape[1]), "The shape of IR1 and IR4 data must be the same"

        # get longitude and latitude
        longitude = IR1['longitude'].to_numpy()
        latitude = IR1['latitude'].to_numpy()
        # meshgrid
        longitude, latitude = np.meshgrid(longitude, latitude)
        # get solar altitude and zenith angle
        solar_altitude, solar_zenith = solar_altitude_zenith(date, longitude, latitude)
        # we only detect seafog at night
        if not (solar_zenith[-1, 0] >= 90 and solar_zenith[-1, -1] >= 90):
            logger.error(f"`night_seafog` can only detect seafog at night. Use `daytime_seafog` to detect diurnal seafog.")
            raise ValueError(f"`night_seafog` can only detect seafog at night. Use `daytime_seafog` to detect diurnal seafog.")
        # create an empty array to store seafog height, with nan means there is no seafog
        seafog = np.full(IR1.shape, np.nan)
        # diff IR4 and IR1 to take out seafog
        diff_IR4_IR1 = IR4.to_numpy() - IR1.to_numpy()

        index = np.logical_and(diff_IR4_IR1 > self.diff_range_night_IR[0],
                               diff_IR4_IR1 < self.diff_range_night_IR[1])
        seafog[index] = 100 * (-2.12 + 1.91 * np.abs(diff_IR4_IR1[index]) / 2)

        # the height of seafog should less than 600
        seafog[seafog > self._maximum_height] = np.nan
        # remove single point
        seafog = remove_single_point(seafog)
        # the seafog detection has been done

        return DataArray(name='seafog', data=seafog, dims=IR1.dims, coords=IR1.coords, attrs={
            'units': 'm',
            'long_name': "height of seafog's top",
            'standard_name': 'seafog',
            'description': f"height of seafog's top, calculated by package seafog at {ctime()}"
        })


detect_seafog = Seafog()

__all__ = ['detect_seafog']



"""
seafog.avhrr contains the function to download AVHRR daily SST (Sea Surface Temperature) data
from the `website <https://coastwatch.pfeg.noaa.gov/erddap/griddap/nceiPH53sstd1day.html>`_
"""

from datetime import datetime, timedelta
from json import loads, dumps
from os import makedirs
from os.path import exists
from time import time
from typing import Callable

from rich.progress import Progress

from .res import AVHRR_CACHE_FOLDER
from .utils import download_url, logger


def avhrr_find_data(date: str, save_path: str, proxy_host: str = None, proxy_port: int = None, progress: Progress = None, headers: dict = None,
                    show_progress=True, callback: Callable = None) -> str:
    """
    download avhrr daily sst data from the `website <https://coastwatch.pfeg.noaa.gov/erddap/griddap/nceiPH53sstd1day.html>`_.

    :param date: UTC time, for example, "2020-01-01 00:00".
    :param save_path: the directory path to store downloaded data.
    :param proxy_host: host address of proxy server.
    :param proxy_port: port number of proxy server.
    :param progress: rich.progress.Progress object to display download progress.
    :param headers: self-defined http headers.
    :param show_progress: if True, display download progress in console.
    :param callback: callable object that will be called during data download.
                     it can be a function or any other Python object.
                     `callback` should accept two params: `total_size` and `step_size`.
    :return: path of downloaded data.

    download data:

    >>> avhrr_find_data("2020-01-01 00:00", "data", show_progress=False) # set show_progress=False will disable the progress bar.
    'data/20200101141457-NCEI-L3C_GHRSST-SSTskin-AVHRR_Pathfinder-PFV5.3_NOAA19_G_2020001_day-v02.0-fv01.0.nc'

    """
    # for example, https://coastwatch.pfeg.noaa.gov/erddap/files/nceiPH53sstd1day/2014/data/
    ROOT_URL = "https://coastwatch.pfeg.noaa.gov/erddap/files/nceiPH53sstd1day/{}/data/"
    # check cache folder
    if not exists(AVHRR_CACHE_FOLDER):
        makedirs(AVHRR_CACHE_FOLDER)
    # parse date
    date = datetime.strptime(date, "%Y-%m-%d %H:%M")
    # read file lists
    cache_file = f"{AVHRR_CACHE_FOLDER}/{date.year}.json"
    # if not exists, download cache
    exist_flag = exists(cache_file)
    # if cache is out of date, refresh it
    refresh_flag = False
    # # store update_time
    update_time = {}
    # # expire time
    expire_time = timedelta(hours=1)
    if not exists(f"{AVHRR_CACHE_FOLDER}/update_time"):
        refresh_flag = True
    else:
        with open(f"{AVHRR_CACHE_FOLDER}/update_time", "r") as f:
            update_time = loads(f.read())
        if str(date.year) not in update_time:
            refresh_flag = True
        elif datetime.fromtimestamp(time()) - datetime.strptime(update_time[str(date.year)], "%Y-%m-%d %H") > expire_time:
            refresh_flag = True
    if not exist_flag or refresh_flag:
        url = ROOT_URL.format(date.year) + ".json"
        logger.debug(f"Download {date.year}'s file list cache")
        status = download_url(url, AVHRR_CACHE_FOLDER, f"{date.year}.json", proxy_host=proxy_host, proxy_port=proxy_port,
                              headers=headers, show_progress=show_progress, progress=progress)
        if status != 200:
            logger.warn(f"Fail to download file lists! Use old file list.")
        else:
            # update update_time
            update_time[str(date.year)] = datetime.fromtimestamp(time()).strftime("%Y-%m-%d %H")
            with open(f"{AVHRR_CACHE_FOLDER}/update_time", "w") as f:
                f.write(dumps(update_time))
    # read file lists from json
    # # check the cache file
    if not exists(cache_file):
        return ""
    with open(cache_file, 'r') as f:
        file_list = loads(f.read())['table']['rows']
    # find the corresponding filename and its size
    filename = None
    size = None
    for file in file_list:
        if file[0][:8] == date.strftime("%Y%m%d"):
            filename = file[0]
            size = file[2]
    # check
    if filename is None:
        logger.error(f"Data at {date.strftime('%Y-%m-%d')} not found")
        return ""
    # download data
    if not exists(save_path):
        makedirs(save_path)
    # if the file exists, return
    if exists(f"{save_path}/{filename}"):
        return f"{save_path}/{filename}"

    url = ROOT_URL.format(date.year) + filename
    # logger.debug(f"Download data from {url}")
    status = download_url(url, save_path, filename, proxy_host=proxy_host, proxy_port=proxy_port, size=size, headers=headers,
                          show_progress=show_progress, progress=progress, callback=callback)
    if status != 200:
        logger.error(f"Fail to download data from {url}")
        return ""

    return f"{save_path}/{filename}"


__all__ = ['avhrr_find_data']



"""
Derived from toolkit ``Plot4Verify_Seafog``, ``seafog`` is a Python package to help download essential data, detect seafog top height, seafog area, and plot images about seafog.
"""

from avhrr import *
from detect import *
from gfs import *
from goos_sst import *
from himawari import *
from pgm import *
from res import LANDMASK_file
from utils import *



"""
seafog.goos_sst provides methods to download NearGOOS data from `JMA <https://www.data.jma.go.jp/gmd/goos/data/pub/JMA-product/>`_ and reads it.
"""

from datetime import datetime
from os import makedirs
from os.path import exists
from time import time
from typing import Tuple, Callable

import numpy as np
from rich.progress import Progress
from xarray import DataArray

from .utils import decompress_file, download_url, logger

# For 0.25 degree GOOS data
ROOT_URL = "https://www.data.jma.go.jp/goos/data/pub/JMA-product/"
# the {} for data year
GOOS_SST_URL = "mgd_sst_glb_D/{}"
# the {} for data date.
# for example, mgd_sst_glb_D20220101.txt.gz
GOOS_RE_SST_NAME_TEMPLATE = "re_mgd_sst_glb_D{}.txt.gz"
GOOS_SST_NAME_TEMPLATE = "mgd_sst_glb_D{}.txt.gz"

# For 0.1 degree High-GOOS data
# the {} for data year
HIGH_GOOS_SST_URL = "him_sst_pac_D/{}"
# the {} for data date.
# for example, him_sst_pac_D20220615.txt
HIGH_GOOS_NAME_TEMPLATE = "him_sst_pac_D{}.txt"


def generate_data_info(date: str, save_path: str, resolution="low", reanalysis=False) -> Tuple[str, str, str]:
    """
    generate data filename, download url, and the save path.
    
    :param date: for example, "2022-05-19 12:00", UTC time
    :param save_path: path to save data.
    :param resolution: ``low``: 0.25 degrees; ``high``: 0.1 degree.
    :param reanalysis: If ``Ture``, use reanalysis name template.
    :return: tuple(filename, download url, save path)
    """
    # generate date time
    date_value = datetime.strptime(date, "%Y-%m-%d %H:%M")
    date = date_value.strftime("%Y%m%d")
    data_year = date[:4]

    # generate filename, url and save path
    if resolution == "low":
        current_time = datetime.fromtimestamp(time())
        logger.debug(f"Current date: {current_time.strftime('%Y-%m-%d %H:%M')}")

        if reanalysis or (current_time.year - date_value.year) >= 3:
            filename = GOOS_RE_SST_NAME_TEMPLATE.format(date)
        else:
            filename = GOOS_SST_NAME_TEMPLATE.format(date)

        url = f"{ROOT_URL}/{GOOS_SST_URL.format(data_year)}/{filename}"
        data_path = f"{save_path}/{filename}"

    elif resolution == "high":
        filename = HIGH_GOOS_NAME_TEMPLATE.format(date)
        url = f"{ROOT_URL}/{HIGH_GOOS_SST_URL.format(data_year)}/{filename}"
        data_path = f"{save_path}/{filename}"

    else:
        logger.error(f"Unknown resolution: {resolution}. Valid values: ['low', 'high']")
        raise ValueError

    return filename, url, data_path


def goos_sst_parser(sst_filename: str, resolution="low") -> DataArray:
    """
    Read NearGOOS sst data.
    
    :param sst_filename: GOOS sst data filename
    :param resolution: `low`: 0.25 degree; `high`: 0.1 degree.
    :return: SST data stored in DataArray, with latitude and longitude and nan value (where is 999 and 88.8)
    """
    # check the file
    assert exists(sst_filename), f"file {sst_filename} does not exist"

    # read all lines in input file
    with open(sst_filename, 'r') as f:
        inputs = f.readlines()
    # the first line is date
    inputs = inputs[1:]

    # parse data, every three characters is a number
    data = []
    for line in inputs:
        # remove the \n at the end of line
        line = line[:-1]
        data.append([int(line[i:i + 3]) for i in range(0, len(line), 3)])

    data = np.asarray(data).astype(float) / 10
    # remove invalid data which is larger than 80
    data[data > 80] = np.nan
    # turn data upside down, convert unit to K
    data = data[::-1, :] + 273.15

    # generate longitude and latitude
    if resolution == "low":
        longitude = np.linspace(0.125, 359.875, 1440)
        latitude = np.linspace(-89.875, 89.875, 720)

    elif resolution == "high":
        longitude = np.linspace(100.05, 179.95, 800)
        latitude = np.linspace(0.05, 59.95, 600)

    else:
        logger.error(f"Unknown resolution: {resolution}. Valid values: ['low', 'high']")
        raise ValueError

    # generate xarray data array
    data = DataArray(
        name='sst', data=data, dims=['latitude', 'longitude'], coords={
            'latitude': latitude,
            'longitude': longitude
        }, attrs={
            'units': 'K',
            "long_name": "Sea surface temperature",
        }
        )

    return data


def goos_sst_find_data(
    date: str, save_path: str, resolution="low", reanalysis=False, proxy_host: str = None, proxy_port: int = None,
    progress: Progress = None, headers: dict = None, show_progress=True, callback: Callable = None
    ) -> str:
    """
    download NearGOOS SST data and save it to the specified path.
    
    :param date: for example, "2022-05-19 12:00", UTC time.
    :param save_path: path to save data.
    :param resolution: `low`: 0.25 degrees; `high`: 0.1 degrees.
    :param reanalysis: If ``True``, download reanalysis data.
    :param proxy_host: http and https proxy server address.
    :param proxy_port: http and https proxy server port.
    :param progress: ``rich.progress.Progress`` object to display progress bar.
    :param headers: self-defined http headers.
    :param show_progress: if True, display download progress in the console.
    :param callback: callable object that will be called every iteration during data download.
                     ``callback`` should accept two params, `total_size` and `step_size`
    :return: data path.
    
    download NearGOOS data:
    
    >>> goos_sst_find_data("2024-07-29 00:00", save_path="data", show_progress=False)
    'data/mgd_sst_glb_D20240729.txt'

    download HighGOOS data:
    
    >>> goos_sst_find_data("2024-07-29 00:00", save_path="data", resolution="high", show_progress=False)
    'data/him_sst_pac_D20240729.txt'

    """
    # check the save path, if it doesn't exist, download data directly
    if not exists(save_path):
        makedirs(save_path)

    # parse date and determine data file name
    data_name, url, data_path = generate_data_info(date, save_path, resolution=resolution, reanalysis=reanalysis)

    # check if the file exists
    if exists(data_path[:-3]):
        return data_path[:-3]

    # download data and decompress it
    # logger.debug(f"Downloading file from {url}")
    code = download_url(
        url, save_path, data_name, proxy_host=proxy_host, proxy_port=proxy_port,
        headers=headers, show_progress=show_progress, progress=progress, callback=callback
        )
    if code == 404:
        if data_name.startswith("re_"):
            logger.error(f"The file {data_name} doesn't exist in the server (status 404)")
            raise FileNotFoundError(f"The file {data_name} doesn't exist in the server (status 404)")
        else:
            logger.warning(f"The file {data_name} doesn't exist in the server, try to download reanalysis version")
            return goos_sst_find_data(
                date=date, save_path=save_path, resolution=resolution, reanalysis=True, proxy_host=proxy_host, proxy_port=proxy_port,
                progress=progress, headers=headers, show_progress=show_progress, callback=callback
            )

    elif code != 200:
        logger.error(f"Failed to download file {data_name}, status code is {code}. May be you can try again later, or check if this url is right: {url}")
        raise ConnectionError

    # No need to decompress for High-GOOS
    if resolution == "low":
        decompress_file(data_path)
        data_path = data_path[:-3]

    return data_path


__all__ = ['goos_sst_parser', 'goos_sst_find_data', 'generate_data_info']



"""
seafog.gfs provides methods to download GFS forecast results from `NOAA <https://nomads.ncep.noaa.gov/cgi-bin/filter_gfs_0p25.pl>`_.
"""

from copy import copy
from datetime import datetime, timedelta
from os import makedirs
from os.path import exists
from time import time
from typing import Dict, Union, Callable, Tuple
from urllib.parse import urlencode

from rich.progress import Progress

from .utils import logger, download_url

# The root URL to download GFS data
ROOT_URL = "https://nomads.ncep.noaa.gov/cgi-bin/filter_gfs_0p25.pl"

# GFS only keeps ten days of data, so we need to set an expiry time and check it
# Units: day
EXPIRY_TIME = 10

# Here is all the params which could be used to download GFS forecast data, which is parsed from the url generated by the website below.
# For more details, see this website: https://nomads.ncep.noaa.gov/gribfilter.php?ds=gfs_0p25
# Select params and click "Show Url", you will see the URL generated by the website.
# Default template dict is suitable to download GFS data for WRF running.
# There are four kinds of key:
# "dir": The store path of forecast data. {DATE} means the date of the forecast, {FORECAST_INDEX} is one of: 00, 06, 12, 18
# "file": From which file you want to collect data.
# "lev_*": Level settings.
# "var_*": Variables settings.
GFS_PARAMS_DICT_TEMPLATE = {
    "dir": "/gfs.{DATE}/{FORECAST_INDEX}/atmos",
    "file": "gfs.t{FORECAST_INDEX}z.pgrb2.0p25.f000",
    "lev_0-0.1_m_below_ground": "on",
    "lev_0.01_mb": "off",
    "lev_0.02_mb": "off",
    "lev_0.04_mb": "off",
    "lev_0.07_mb": "off",
    "lev_0.1-0.4_m_below_ground": "on",
    "lev_0.1_mb": "off",
    "lev_0.2_mb": "off",
    "lev_0.33-1_sigma_layer": "off",
    "lev_0.4-1_m_below_ground": "on",
    "lev_0.44-0.72_sigma_layer": "off",
    "lev_0.44-1_sigma_layer": "off",
    "lev_0.4_mb": "off",
    "lev_0.72-0.94_sigma_layer": "off",
    "lev_0.7_mb": "off",
    "lev_0.995_sigma_level": "off",
    "lev_0C_isotherm": "off",
    "lev_1-2_m_below_ground": "on",
    "lev_1000_m_above_ground": "off",
    "lev_1000_mb": "on",
    "lev_100_m_above_ground": "off",
    "lev_100_mb": "on",
    "lev_10_m_above_ground": "on",
    "lev_10_m_above_mean_sea_level": "off",
    "lev_10_mb": "off",
    "lev_150_mb": "on",
    "lev_15_mb": "off",
    "lev_180-0_mb_above_ground": "off",
    "lev_1829_m_above_mean_sea_level": "off",
    "lev_1_hybrid_level": "off",
    "lev_1_mb": "off",
    "lev_200_mb": "on",
    "lev_20_m_above_ground": "off",
    "lev_20_mb": "off",
    "lev_250_mb": "on",
    "lev_255-0_mb_above_ground": "off",
    "lev_2743_m_above_mean_sea_level": "off",
    "lev_2_hybrid_level": "off",
    "lev_2_m_above_ground": "on",
    "lev_2_mb": "off",
    "lev_30-0_mb_above_ground": "off",
    "lev_3000-0_m_above_ground": "off",
    "lev_300_mb": "on",
    "lev_30_m_above_ground": "off",
    "lev_30_mb": "off",
    "lev_350_mb": "on",
    "lev_3658_m_above_mean_sea_level": "off",
    "lev_3_mb": "off",
    "lev_4000_m_above_ground": "off",
    "lev_400_mb": "on",
    "lev_40_m_above_ground": "off",
    "lev_40_mb": "off",
    "lev_450_mb": "on",
    "lev_500_mb": "on",
    "lev_50_m_above_ground": "off",
    "lev_50_mb": "on",
    "lev_550_mb": "on",
    "lev_5_mb": "off",
    "lev_6000-0_m_above_ground": "off",
    "lev_600_mb": "on",
    "lev_650_mb": "on",
    "lev_700_mb": "on",
    "lev_70_mb": "on",
    "lev_750_mb": "on",
    "lev_7_mb": "off",
    "lev_800_mb": "on",
    "lev_80_m_above_ground": "off",
    "lev_850_mb": "on",
    "lev_90-0_mb_above_ground": "off",
    "lev_900_mb": "on",
    "lev_925_mb": "on",
    "lev_950_mb": "on",
    "lev_975_mb": "on",
    "lev_PV=-2e-06_(Km^2/kg/s)_surface": "off",
    "lev_PV=2e-06_(Km^2/kg/s)_surface": "off",
    "lev_boundary_layer_cloud_layer": "off",
    "lev_cloud_ceiling": "off",
    "lev_convective_cloud_bottom_level": "off",
    "lev_convective_cloud_layer": "off",
    "lev_convective_cloud_top_level": "off",
    "lev_entire_atmosphere": "off",
    "lev_entire_atmosphere_(considered_as_a_single_layer)": "off",
    "lev_high_cloud_bottom_level": "off",
    "lev_high_cloud_layer": "off",
    "lev_high_cloud_top_level": "off",
    "lev_highest_tropospheric_freezing_level": "off",
    "lev_low_cloud_bottom_level": "off",
    "lev_low_cloud_layer": "off",
    "lev_low_cloud_top_level": "off",
    "lev_max_wind": "off",
    "lev_mean_sea_level": "on",
    "lev_middle_cloud_bottom_level": "off",
    "lev_middle_cloud_layer": "off",
    "lev_middle_cloud_top_level": "off",
    "lev_planetary_boundary_layer": "off",
    "lev_surface": "on",
    "lev_top_of_atmosphere": "off",
    "lev_tropopause": "off",
    "var_4LFTX": "off",
    "var_ABSV": "off",
    "var_ACPCP": "off",
    "var_ALBDO": "off",
    "var_APCP": "off",
    "var_CAPE": "off",
    "var_CFRZR": "off",
    "var_CICEP": "off",
    "var_CIN": "off",
    "var_CLWMR": "off",
    "var_CNWAT": "off",
    "var_CPOFP": "off",
    "var_CPRAT": "off",
    "var_CRAIN": "off",
    "var_CSNOW": "off",
    "var_CWAT": "off",
    "var_CWORK": "off",
    "var_DLWRF": "off",
    "var_DPT": "off",
    "var_DSWRF": "off",
    "var_DZDT": "off",
    "var_FLDCP": "off",
    "var_FRICV": "off",
    "var_GFLUX": "off",
    "var_GRLE": "off",
    "var_GUST": "off",
    "var_HCDC": "off",
    "var_HGT": "on",
    "var_HINDEX": "off",
    "var_HLCY": "off",
    "var_HPBL": "off",
    "var_ICAHT": "off",
    "var_ICEC": "on",
    "var_ICEG": "off",
    "var_ICETK": "off",
    "var_ICETMP": "off",
    "var_ICMR": "off",
    "var_LAND": "on",
    "var_LCDC": "off",
    "var_LFTX": "off",
    "var_LHTFL": "off",
    "var_MCDC": "off",
    "var_MSLET": "off",
    "var_O3MR": "off",
    "var_PEVPR": "off",
    "var_PLPL": "off",
    "var_POT": "off",
    "var_PRATE": "off",
    "var_PRES": "on",
    "var_PRMSL": "on",
    "var_PWAT": "off",
    "var_REFC": "off",
    "var_REFD": "off",
    "var_RH": "on",
    "var_RWMR": "off",
    "var_SFCR": "off",
    "var_SHTFL": "off",
    "var_SNMR": "off",
    "var_SNOD": "off",
    "var_SOILL": "off",
    "var_SOILW": "on",
    "var_SOTYP": "off",
    "var_SPFH": "off",
    "var_SUNSD": "off",
    "var_TCDC": "off",
    "var_TMAX": "off",
    "var_TMIN": "off",
    "var_TMP": "on",
    "var_TOZNE": "off",
    "var_TSOIL": "on",
    "var_U-GWD": "off",
    "var_UFLX": "off",
    "var_UGRD": "on",
    "var_ULWRF": "off",
    "var_USTM": "off",
    "var_USWRF": "off",
    "var_V-GWD": "off",
    "var_VEG": "off",
    "var_VFLX": "off",
    "var_VGRD": "on",
    "var_VIS": "off",
    "var_VRATE": "off",
    "var_VSTM": "off",
    "var_VVEL": "off",
    "var_VWSH": "off",
    "var_WATR": "off",
    "var_WEASD": "on",
    "var_WILT": "off",
    "leftlon": "70",
    "rightlon": "165",
    "toplat": "65",
    "bottomlat": "8"
}


def get_gfs_param_template(date: Union[str, datetime], forecast_index: int) -> Dict[str, str]:
    """
    Get param template dict to download GFS data.

    :param date: Date of the GFS data you download, either a string or datetime object.
                 For example, ``"2024-04-02"``.
    :param forecast_index: GFS runs 4 times a day, so there will be four indexes, [0, 6, 12, 18].
    :return: Parameters dict to download GFS data.
    """
    global GFS_PARAMS_DICT_TEMPLATE

    # check the date form
    if isinstance(date, str):
        date = datetime.strptime(date, "%Y-%m-%d")

    # get today's date
    today_date = datetime.fromtimestamp(time())

    # check if `date` is ten days before today
    if today_date - date > timedelta(days=10):
        logger.error(f"You can't download GFS data more than 10 days before, because GFS only keeps 10 days results")
        raise ValueError

    # check forecast index
    if forecast_index not in [0, 6, 12, 18]:
        logger.error(f"`forecast_index` should be one of [0, 6, 12, 18]")
        raise ValueError

    # copy template
    new_template = copy(GFS_PARAMS_DICT_TEMPLATE)

    # convert forecast_index and date to string
    forecast_index = str(forecast_index).rjust(2, '0')
    date = date.strftime("%Y%m%d")

    # fill value
    new_template["dir"] = new_template["dir"].format(DATE=date, FORECAST_INDEX=forecast_index)
    new_template["file"] = new_template["file"].format(FORECAST_INDEX=forecast_index)

    return new_template


def generate_gfs_download_url(param_dict: Dict[str, str]) -> str:
    """
    Generate GFS download URL based on the parameters' dict.

    :param param_dict: Parameters dict.
    :return: URL to download GFS data.

    Generate default url from template:

    >>> data_date = datetime.fromtimestamp(time()).strftime("%Y-%m-%d")
    >>> default_params = get_gfs_param_template(data_date, forecast_index=0)
    >>> url = generate_gfs_download_url(default_params)

    """
    # we need to remove the lev_* and var_* key with value "off"
    new_param_dict = {}
    for key in param_dict:

        if key.startswith("lev_") or key.startswith("var_"):
            if param_dict[key] != "off":
                new_param_dict[key] = param_dict[key]
        else:
            new_param_dict[key] = param_dict[key]

    # generate URL
    param_string = urlencode(new_param_dict)

    return f"{ROOT_URL}?{param_string}"


def gfs_find_data(date: str, save_path: str, area: Tuple[int, int, int, int], forecast_index: int, param_dict: Dict[str, str] = None,
                  proxy_host: str = None, proxy_port: int = None, progress: Progress = None, show_progress=True, callback: Callable = None) -> str:
    """
    Download GFS data in the last 10 days from `NOAA <https://nomads.ncep.noaa.gov/cgi-bin/filter_gfs_0p25.pl>`_.
    NOAA only keeps the last 10-day results, any date earlier than 10 days will raise an ValueError.

    :param date: Date of GFS data you want to download.
                 For example, ``"2024-04-02"``.
    :param save_path: Save path of GFS data.
    :param area: Longitude and latitude range, for example, (lon1, lon2, lat1, lat2).
    :param forecast_index: GFS runs 4 times a day, so there will be four indexes, [0, 6, 12, 18].
    :param param_dict: You can provide custom parameter dict to download GFS data; otherwise, we will download with default selection of variables.
    :param proxy_host: Http and https proxy server address.
    :param proxy_port: Http and https proxy server port.
    :param progress:  Living progress object.
    :param show_progress: If True, display download progress in console.
    :param callback: Callable object that will be called every iteration during data download.
                     ``callback`` should accept two params, `total_size` and `step_size`
    :return: path of downloaded data.

    Download today's data in area (116E-128E, 30N-40N) with ``forecast_index=0`` and default params:

    >>> today_date = datetime.fromtimestamp(time()).strftime("%Y-%m-%d")
    >>> gfs_find_data(today_date, "data", area=(116, 128, 30, 40), forecast_index=0, show_progress=False)
    'data/gfs.t00z.pgrb2.0p25.f000'
    
    Download data with custom params:

    >>> today_date = datetime.fromtimestamp(time()).strftime("%Y-%m-%d")
    >>> params = get_gfs_param_template(today_date, forecast_index=0)
    >>> # change any params
    >>> data_path = gfs_find_data(today_date, "data", area=(116, 128, 30, 40), forecast_index=0, show_progress=False)

    """
    # check the save path
    if not exists(save_path):
        makedirs(save_path)

    # get template parameters dict
    if param_dict is None:
        param_dict = get_gfs_param_template(date, forecast_index)

    # update longitude and latitude range
    lon1, lon2, lat1, lat2 = area
    param_dict["leftlon"] = str(lon1)
    param_dict["rightlon"] = str(lon2)
    param_dict["bottomlat"] = str(lat1)
    param_dict["toplat"] = str(lat2)

    # generate save name.
    # data of the same date may have different variables and in different areas, so we just use value `param_dict['file']` as the save name.
    save_name = param_dict["file"]

    # generate URL
    url = generate_gfs_download_url(param_dict)

    # download
    return_code = download_url(url, save_path, save_name, proxy_host=proxy_host, proxy_port=proxy_port, show_progress=show_progress, progress=progress, callback=callback)

    # check
    if return_code != 200:
        logger.error(f"Fail to download file {save_name}, status code is {return_code}")
        raise ConnectionError

    else:
        return f"{save_path}/{save_name}"


__all__ = ["get_gfs_param_template", "generate_gfs_download_url", "gfs_find_data"]



"""
This Python file contains some utility functions for the project.
"""

import gzip
import logging
import sys
import warnings
from datetime import datetime
from multiprocessing import Pipe, Process
from multiprocessing.connection import Connection
from os import cpu_count, listdir, makedirs, remove, rename
from os.path import abspath, basename, dirname, exists, getsize
from shutil import unpack_archive
from time import sleep
from typing import Callable, Optional, Tuple, Union

import ephem
import numpy as np
import pycurl
import xarray as xr
from global_land_mask import is_land
from requests import get
from requests.exceptions import ConnectTimeout, ProxyError, RequestException, SSLError
from rich.logging import RichHandler
from rich.progress import BarColumn, DownloadColumn, Progress, TextColumn, TimeRemainingColumn, TransferSpeedColumn

# close numpy RuntimeWarning
warnings.filterwarnings("ignore")

# init a logger
logger = logging.getLogger("seafog")
formatter = logging.Formatter("%(name)s :: %(message)s", datefmt="%m-%d %H:%M:%S")
# use rich handler
handler = RichHandler()
handler.setFormatter(formatter)
logger.addHandler(handler)
logger.setLevel(logging.WARNING)
# remove handler in pgm reader and set handler
_logger = logging.getLogger("pgm reader")
for _handler in _logger.handlers:
    _logger.removeHandler(_handler)
_logger.addHandler(handler)

# define progress
PROGRESS = Progress(
    TextColumn("{task.description}"),
    BarColumn(bar_width=None),
    "[progress.percentage]{task.percentage:>3.1f}%",
    "•", DownloadColumn(),
    "•", TransferSpeedColumn(),
    "•", TimeRemainingColumn()
)

sun = ephem.Sun()  # type: ignore
observer = ephem.Observer()

# disguise requests.get as wget
HEADERS = {
    "User-Agent": "Wget/1.12 (linux-gnu)"
}


def _multi_process_env_check() -> bool:
    """
    On Windows and macOS, we need something like ``if __name__ == "__main__"`` to avoid the RuntimeError.
    So check the main entry file first.

    :return:
    :rtype:
    """
    if sys.platform == "linux":
        return True

    main_entry_file = sys.argv[0]
    with open(main_entry_file, 'r') as f:
        contents = f.read()

    if "if __name__ == '__main__':" not in contents or 'if __name__ == "__main__"' not in contents:
        return False
    else:
        return True


class CurlSingleProcessCallback:
    """
    This callback is used to update progress when download the data in the single process.

    """

    def __init__(self, curl_handle: pycurl.Curl, progress: Progress = None, filename: str = "", callback: Callable = None):
        """
        A callback class that will be used to trace the progress of FTP download.

        :param curl_handle: The curl object to download files.
                            We need to access this to stop downloading.
        :param progress: A Progress object to display download progress
        :param filename: Save filename.
        :param callback: Callable object that will be called every iteration of `res.iter_content`.
        `callback` should accept two params: `total_size` and `step_size`.
        `callback` can return a non-zero and non-None value to stop download.
        """
        self.curl = curl_handle
        self.progress: Progress = progress
        self.task_id = None
        self.callback: Callable = callback
        self.filename = filename
        self.last_downloaded_total = 0
        self.start_size = 0

    def set_progress(self, progress: Progress, filename=""):
        """
        Set progress info.
        :param progress:
        :type progress:
        :param filename:
        :type filename:
        :return:
        :rtype:
        """
        self.progress = progress
        self.filename = filename

    def set_start_size(self, size: int):
        """
        If we continue to download from .part file, we need to tell progress the size of download part.

        :param size: Download part's size.
        :return:
        """
        self.start_size = size

    def __call__(self, download_total: int, downloaded_total: int, upload_total: int, uploaded_total: int):
        # Check download_total and downloaded_total
        # If both are 0, download doesn't begin
        if download_total == 0 and downloaded_total == 0:
            return
        # Transform to kb
        download_total = download_total
        downloaded_total = downloaded_total
        # Calculate step size
        step_size = downloaded_total - self.last_downloaded_total
        self.last_downloaded_total = downloaded_total

        # Check task id. If it's None, create a new task
        if self.task_id is None and isinstance(self.progress, Progress):
            # create a task
            self.task_id = self.progress.add_task(f"[red]{self.filename}[red]", total=download_total)
            # set start size
            self.progress.update(self.task_id, advance=self.start_size)
            # update progress
            self.progress.update(self.task_id, advance=step_size)
        elif self.task_id is not None:
            self.progress.update(self.task_id, advance=step_size)
        if self.callback is not None:
            return_code = self.callback(download_total, step_size)
            # check return code
            if return_code is not None and return_code != 0:
                logger.info(f"Stop downloading by callback")
                # pause at first
                self.curl.pause(pycurl.PAUSE_ALL)
                # just return non-zero value can stop downloading
                return return_code


class CurlMultiProcessCallback:
    """
    This callback is used to send download progress to the main progress when download the file in multiple processes.

    """

    def __init__(self, send_pipe: Connection, downloaded_size=0, restart=False):
        self.send_pipe = send_pipe
        self.last_downloaded = downloaded_size
        if downloaded_size != 0 and not restart:
            self.send_pipe.send(downloaded_size)

    def __call__(self, download_total: int, downloaded_total: int, upload_total: int, uploaded_total: int):
        if download_total == 0 and downloaded_total == 0:
            return
        else:
            step_size = downloaded_total - self.last_downloaded
            self.last_downloaded = downloaded_total
            self.send_pipe.send(step_size)


class CurlDownloader:
    """
    A pycurl wrapper class to download files in single or multiple processes.

    """

    def __init__(self, url: str, save_path: str, filename: str, proxy_host: Optional[str] = None, proxy_port: Optional[int] = None, timeout=720, connection_timeout=30):
        self.url = url
        self.save_path = save_path
        self.filename = filename
        self.proxy_host = proxy_host
        self.proxy_port = proxy_port
        self.timeout = timeout
        self.connection_timeout = connection_timeout

    def multiple_process(self, part_filename: str, start: int, end: int, send_pipe: Connection, force_download: bool):
        """
        This function should be called in another process, and it is used to download part of the file.
        It can automatically restart the task up to five times if download is failed.

        :param part_filename: Filename of the temporal file.
        :type part_filename: str
        :param start: Start point to download.
        :type start: int
        :param end: End point of the part.
        :type end: int
        :param send_pipe: The pipe to send downloaded size.
        :type send_pipe: Connection
        :param force_download: If ignore the existing data and re-download it.
        :type force_download: bool
        :return:
        :rtype:
        """
        try_time = 1
        success_flag = 0
        restart = False

        while try_time <= 5 and not success_flag:
            try:
                self._multiple_process(part_filename, start, end, send_pipe, force_download, restart)
                success_flag = 1
            except pycurl.error as e:
                restart = True
                error_code = e.args[0]
                if error_code in [23, 42]:
                    logger.error(f"Stop download by the user")
                    try_time = 5
                else:
                    logger.warning(f"Subprocess exits because of '{e.args[1]}' (error code {e.args[0]}), retry ({try_time} times)")

            try_time += 1

        if not success_flag:
            logger.error(f"Failed to download the file")
            raise ConnectionError

    def _multiple_process(self, part_filename: str, start: int, end: int, send_pipe: Connection, force_download: bool, restart=False):
        """
        This function should be called in another process, and it is used to download part of the file.

        :param part_filename: Filename of the temporal file.
        :type part_filename: str
        :param start: Start point to download.
        :type start: int
        :param end: End point of the part.
        :type end: int
        :param send_pipe: The pipe to send downloaded size.
        :type send_pipe: Connection
        :param force_download: If ignore the existing data and re-download it.
        :type force_download: bool
        :param restart: If True, doesn't report the downloaded file size.
        :type restart: bool
        :return: True if download successfully, else False.
        :rtype: bool
        """
        # curl settings
        curl = pycurl.Curl()
        curl.setopt(pycurl.CONNECTTIMEOUT, self.connection_timeout)
        if self.timeout >= 0:
            curl.setopt(pycurl.TIMEOUT, self.timeout)
        curl.setopt(pycurl.URL, self.url)
        curl.setopt(pycurl.NOPROGRESS, False)
        if self.proxy_host is not None and self.proxy_port is not None:
            if not self.proxy_host.startswith("socks5"):
                logger.error(f"Only support `socks5` protocol")
                raise ValueError
            else:
                curl.setopt(pycurl.PROXY, self.proxy_host)
                curl.setopt(pycurl.PROXYPORT, self.proxy_port)
                curl.setopt(pycurl.PROXYTYPE, pycurl.PROXYTYPE_SOCKS5)

        save_path = f"{self.save_path}/{part_filename}"
        if not force_download and exists(save_path):
            downloaded_size = getsize(save_path)
            if downloaded_size > (end - start):
                open_type = "wb"
            else:
                open_type = "ab"
                start = start + downloaded_size
        else:
            downloaded_size = 0
            open_type = "wb"

        curl.setopt(pycurl.RANGE, f"{start}-{end}")
        curl.setopt(pycurl.XFERINFOFUNCTION, CurlMultiProcessCallback(send_pipe, downloaded_size, restart))

        with open(save_path, open_type) as f:
            curl.setopt(pycurl.WRITEDATA, f)
            curl.perform()

    def single_process(self, progress: Progress = None, callback: Callable = None, force_download=False) -> bool:
        """
        This function can be called directly to download file.
        It can automatically restart the task up to 5 times if download is failed.

        :param progress: 
        :type progress: Progress
        :param callback: 
        :type callback: Callable
        :param force_download: If ignore the existing data and re-download it.
        :type force_download: bool
        :return: True if download successfully, else False.
        :rtype: bool
        """
        try_time = 1
        success_flag = 0
        curl = pycurl.Curl()

        while try_time <= 5 and not success_flag:
            curl.reset()
            curl.setopt(pycurl.CONNECTTIMEOUT, self.connection_timeout)
            if self.timeout >= 0:
                curl.setopt(pycurl.TIMEOUT, self.timeout)
            curl.setopt(pycurl.URL, self.url)
            curl.setopt(pycurl.NOPROGRESS, False)
            progress_callback = CurlSingleProcessCallback(curl_handle=curl, progress=progress, filename=self.filename, callback=callback)
            curl.setopt(pycurl.XFERINFOFUNCTION, progress_callback)

            if self.proxy_host is not None and self.proxy_port is not None:
                if not self.proxy_host.startswith("socks5"):
                    logger.error(f"Only support `socks5` protocol")
                    raise ValueError
                else:
                    curl.setopt(pycurl.PROXY, self.proxy_host)
                    curl.setopt(pycurl.PROXYPORT, self.proxy_port)
                    curl.setopt(pycurl.PROXYTYPE, pycurl.PROXYTYPE_SOCKS5)

            save_path = f"{self.save_path}/{self.filename}"
            temp_save_path = f"{save_path}.part"
            if not force_download and exists(temp_save_path):
                downloaded_size = getsize(temp_save_path)
                curl.setopt(pycurl.RESUME_FROM, downloaded_size)
                progress_callback.set_start_size(downloaded_size)
                open_type = "ab"
            else:
                open_type = "wb"

            with open(temp_save_path, open_type) as f:
                curl.setopt(pycurl.WRITEDATA, f)

                try:
                    curl.perform()
                    success_flag = 1
                except pycurl.error as e:
                    error_code = e.args[0]
                    # pycurl's error is ugly
                    if error_code in [23, 42]:
                        logger.error(f"Stop download by the user")
                        try_time = 5
                    else:
                        logger.warning(f"Failed to download because of {e}, retry ({try_time + 1} times)")

            if success_flag:
                rename(temp_save_path, save_path)

            try_time += 1

        if success_flag:
            return True
        else:
            return False

    def start(self, process_num=1, progress: Progress = None, callback: Callable = None, force_download=False) -> bool:
        """
        Start to download the file in the single process or multiple processes.

        :param process_num: The number of processes used to download the file.
        :type process_num: int
        :param progress:
        :type progress: Progress
        :param callback:
        :type callback: Callable
        :param force_download: If ignore the existing data and re-download it.
        :type force_download: bool
        :return: True if download successfully, else False.
        :rtype: bool
        """
        if process_num < 2:
            return self.single_process(progress=progress, callback=callback)

        else:
            if not _multi_process_env_check():
                logger.error(
                    "If you are about to download the file with multiple processes, you need to wrap your code in the block: `if __name__ == '__main__':` "
                    "to avoid the `RuntimeError` from the multiprocessing package. For more information, please check the official documentation: "
                    "https://docs.python.org/3/library/multiprocessing.html#the-process-class"
                )
                raise RuntimeError(
                    "If you are about to download the file with multiple processes, you need to wrap your code in the block: `if __name__ == '__main__':` "
                    "to avoid the `RuntimeError` from the multiprocessing package. For more information, please check the official documentation: "
                    "https://docs.python.org/3/library/multiprocessing.html#the-process-class"
                )

            logger.info(f"You are downloading the file in Multi-Process mode")
            cpu_cores = cpu_count()
            if process_num > cpu_cores:
                process_num = cpu_cores - 2
                if process_num <= 0:
                    process_num = 1
                logger.warning(f"You machine has {cpu_cores} CPUs, so reduce process number to {process_num}")

            part_filenames = [f"{self.filename}.part.{i}" for i in range(process_num)]

            # check the old cache file
            res = []
            for _cache_file in part_filenames:
                if exists(f"{self.save_path}/{_cache_file}"):
                    res.append(True)
                else:
                    res.append(False)

            if exists(f"{self.save_path}/{self.filename}.part.{process_num}"):
                res.append(False)
            else:
                res.append(True)

            if True in res and False in res:
                logger.warning(f"Process number has changed after the last running, clear old cache file for the fresh download")
                cache_file_list = [x for x in listdir(self.save_path) if x.startswith(f"{self.filename}.part")]
                for _cache_file in cache_file_list:
                    remove(f"{self.save_path}/{_cache_file}")

            # get file size
            curl = pycurl.Curl()
            curl.setopt(pycurl.CONNECTTIMEOUT, self.connection_timeout)
            if self.timeout >= 0:
                curl.setopt(pycurl.TIMEOUT, self.timeout)
            curl.setopt(pycurl.URL, self.url)
            curl.setopt(pycurl.NOPROGRESS, True)
            curl.setopt(pycurl.NOBODY, 1)
            curl.perform()
            file_size = int(curl.getinfo(pycurl.CONTENT_LENGTH_DOWNLOAD))

            if isinstance(progress, Progress):
                pid = progress.add_task(f"[red]{self.filename}[red]", total=file_size)
            else:
                pid = None

            chunk_size = file_size // process_num
            process_pool = []
            recv_pipe, send_pipe = Pipe(False)

            for (i, _part_filename) in enumerate(part_filenames[:-1]):
                process_pool.append(
                    Process(
                        target=self.multiple_process, args=[_part_filename, chunk_size * i, chunk_size * (i + 1) - 1, send_pipe, force_download]
                    )
                )
                process_pool[i].start()

            process_pool.append(
                Process(
                    target=self.multiple_process, args=[part_filenames[-1], chunk_size * (process_num - 1), file_size, send_pipe, force_download]
                )
            )
            process_pool[-1].start()

            loop_flag = 1
            stop_flag = 0
            while loop_flag:
                status_code = []
                for _process in process_pool:
                    # update progress here
                    if pid is not None:
                        if recv_pipe.poll():
                            step_size = recv_pipe.recv()
                            progress.update(pid, advance=step_size)
                            # call user callback function
                            if callback is not None:
                                return_code = callback(file_size, step_size)
                                if return_code != 0:
                                    raise KeyboardInterrupt

                    # 0: the process is alive
                    # 1: the process exits abnormally
                    # 2: the process exits normally
                    if _process.is_alive():
                        status_code.append(0)
                    elif _process.exitcode != 0:
                        status_code.append(1)
                    else:
                        status_code.append(2)

                if 1 in status_code:
                    loop_flag = 0
                    stop_flag = 1
                elif 0 not in status_code:
                    loop_flag = 0

            if stop_flag:
                for _process in process_pool:
                    if _process.is_alive():
                        _process.terminate()

                return False
            else:
                # concat files
                temp_file = f"{self.save_path}/{self.filename}.part"
                with open(temp_file, "wb") as f:
                    for _file in part_filenames:
                        with open(f"{self.save_path}/{_file}", "rb") as f2:
                            f.write(f2.read())
                        remove(f"{self.save_path}/{_file}")

                rename(temp_file, f"{self.save_path}/{self.filename}")

                return True


def solar_altitude_zenith_formula(
    date: str,
    longitude: Union[float, np.ndarray],
    latitude: Union[float, np.ndarray]
) -> Union[Tuple[float, float], Tuple[np.ndarray, np.ndarray]]:
    """
    calculate the solar altitude and zenith angle with math formula.

    :param date: date, for example, '2020-05-29 09:13', UTC time
    :param longitude: longitudes. single value or numpy array. units: degree
    :param latitude: latitudes. single value or numpy array. units: degree
    :return: solar altitude and zenith angle. units: degree
    """
    # calculate the hour angle
    hour = datetime.strptime(date, '%Y-%m-%d %H:%M').hour
    minute = datetime.strptime(date, '%Y-%m-%d %H:%M').minute
    hour = hour + minute / 60 - 12
    hour_angle = hour * 180 / 12 + longitude
    # print(hour_angle)

    # calculate the solar declination
    day = datetime.strptime(date, '%Y-%m-%d %H:%M').strftime("%j")
    day = int(day)
    solar_declination = -23.44 * np.cos(np.deg2rad(360 / 365) * (day + 10))
    # print(solar_declination)

    # calculate the solar altitude
    res = np.cos(np.deg2rad(hour_angle)) * np.cos(np.deg2rad(solar_declination)) * np.cos(np.deg2rad(latitude)) + \
          np.sin(np.deg2rad(solar_declination)) * np.sin(np.deg2rad(latitude))

    return np.rad2deg(np.arcsin(res)), np.rad2deg(np.arccos(res))


def solar_altitude_zenith_ephem(
    date: str,
    longitude: Union[float, np.ndarray],
    latitude: Union[float, np.ndarray]
) -> Union[Tuple[float, float], Tuple[np.ndarray, np.ndarray]]:
    """
    calculate the solar altitude and zenith angle with ``ephem`` library.

    :param date: date, for example, '2020-05-29 09:13', UTC time
    :param longitude: longitudes. single value or numpy array. units: degree
    :param latitude: latitudes. single value or numpy array. units: degree
    :return: solar altitude and zenith angle. units: degree
    """
    # transform date
    date = datetime.strptime(date, '%Y-%m-%d %H:%M').strftime("%Y/%m/%d %H:%M:%S")
    # transform from degree to rad
    longitude = np.deg2rad(longitude)
    latitude = np.deg2rad(latitude)
    # check if is an array or a single value
    if isinstance(longitude, np.ndarray) or isinstance(latitude, np.ndarray):
        shape = longitude.shape
        # transform N-D array to 1-D
        longitude = longitude.flatten()
        latitude = latitude.flatten()
        result = [list(_solar_altitude_zenith_ephem(date, lon, lat)) for lon, lat in zip(longitude, latitude)]
        result = np.asarray(result)
        # # transform 1-D array back to N-D
        return result[:, 0].reshape(shape), result[:, 1].reshape(shape)
    else:
        return _solar_altitude_zenith_ephem(date, longitude, latitude)


def _solar_altitude_zenith_ephem(date: str, longitude: float, latitude: float) -> Tuple[float, float]:
    """
    calculate the solar altitude and zenith angle with ``ephem`` library.

    :param date: date, for example, '2020-05-29 09:13', UTC time
    :param longitude: longitude. single value, units: rad
    :param latitude: latitude. single value, units: rad
    :return: solar altitude and zenith angle. units: degree
    """
    # use global object
    global sun
    global observer

    observer.lon = longitude
    observer.lat = latitude
    observer.date = date
    sun.compute(observer)

    return sun.alt / ephem.degree, 90 - sun.alt / ephem.degree


def solar_altitude_zenith(
    date: str,
    longitude: Union[float, np.ndarray],
    latitude: Union[float, np.ndarray],
    method='formula'
) -> Union[Tuple[float, float], Tuple[np.ndarray, np.ndarray]]:
    """
    calculate the solar altitude and zenith angle based on the date, longitude and latitude.

    :param date: UTC date, for example, '2020-05-29 09:13'
    :param longitude: longitude for a single point or an array. units: degree
    :param latitude: latitude for a single point or an array. units: degree
    :param method: method for calculating solar altitude and zenith angle, including 'formula' and 'ephem'. Defaults to 'formula'.
    :return: solar altitude and zenith angle. units: degree
    """
    # if input is ndarray, calculate multiple points
    if method == 'formula':
        return solar_altitude_zenith_formula(date, longitude, latitude)
    elif method == 'ephem':
        return solar_altitude_zenith_ephem(date, longitude, latitude)
    else:
        raise Exception(f"Unknown method: {method}")


def atmosphere_visibility(temperature: Union[float, np.ndarray], solar_zenith: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
    """
    calculate the atmosphere visibility based on the temperature and solar zenith angle.

    :param temperature: temperature, units: K
    :param solar_zenith: solar zenith, units: degree
    :return: atmosphere visibility, units: meter
    """
    # define constant variables
    beta = 0.0685
    sigma = 0.02

    return 45 * np.power(
        (1 - temperature / 100) * beta / (temperature / 100 * np.cos(np.deg2rad(solar_zenith))),
        1 / 3
    ) * np.log(1 / sigma)


def seafog_top_height(temperature: Union[float, np.ndarray], solar_zenith: Union[float, np.ndarray], beta: float = 0.0685) -> Union[float, np.ndarray]:
    """
    Calculate the seafog thickness based on the brightness temperature of the VIS band and solar zenith angle.
    This function only works at daytime.

    :param temperature: Brightness temperature of the VIS band. Units: K.
    :param solar_zenith: Solar zenith angle. Units: degree
    :param beta: Backscatter coefficient of the atmosphere.
    :return: Height of the marine fog top. Units: meter
    """

    return 45 * np.power(
        temperature / 100 * np.cos(np.deg2rad(solar_zenith)) / ((1 - temperature / 100) * beta),
        2 / 3
    )


def decompress_file(file_path: str, save_path: str = None, remove_raw: bool = True, file_format="gz"):
    """
    decompress a file.

    :param remove_raw: Whether to remove raw compressed files.
    :param file_path: Compressed file path.
    :param save_path: The save path of decompressed files, pass None to save the file in the same directory as the compressed file. Default is None
    :param file_format: File format of the compressed file.
    :return:
    """
    # get save path
    if save_path is not None and not exists(save_path):
        makedirs(save_path)
    elif save_path is None:
        save_path = abspath(dirname(file_path))

    if save_path[-1] != '/':
        save_path += '/'

    if file_format == "gz":
        # get filename
        filename = basename(file_path).replace(".gz", "")

        # decompress file
        # unpack_archive(file_path, save_path, format="")
        g_file = gzip.GzipFile(file_path)
        with open(save_path + filename, 'wb') as f:
            f.write(g_file.read())
        g_file.close()
    else:
        unpack_archive(file_path, save_path, format=file_format)

    # remove .gz file
    if remove_raw:
        remove(file_path)


def download_url(
    url: str, save_path: str, filename: str, proxy_host: str = None, proxy_port: int = None, size: int = None, headers: dict = None,
    show_progress=True, progress: Progress = None, callback: Callable = None
) -> int:
    """
    download the file from the url.

    :param url: file url
    :param save_path: save path
    :param filename: save filename.
    :param proxy_host: proxy server address
    :param proxy_port: proxy server port
    :param size: file size in kb. some websites may not return the content size, in which case you should give size to make progress show correctly.
    :param headers: Self-defined http headers.
    :param show_progress: If True, display download progress in console
    :param progress: Progress object to display download progress
    :param callback: Callable object that will be called every iteration of `res.iter_content`. `callback` should accept two params: `total_size` and `step_size`.
                     `callback` can return a non-zero and non-None value to stop download.
    :return: res status code
    """
    if show_progress:
        if isinstance(progress, Progress):
            return _download_url(
                url, save_path, filename, proxy_host=proxy_host, proxy_port=proxy_port, size=size, headers=headers,
                progress=progress, callback=callback
            )
        else:
            with PROGRESS:
                # remove finished task
                for task in PROGRESS.task_ids:
                    PROGRESS.remove_task(task)
                return _download_url(
                    url, save_path, filename, proxy_host=proxy_host, proxy_port=proxy_port, size=size, headers=headers,
                    progress=PROGRESS, callback=callback
                )
    else:
        return _download_url(url, save_path, filename, proxy_host=proxy_host, proxy_port=proxy_port, size=size, headers=headers, callback=callback)


def _download_url(
    url: str, save_path: str, filename: str, proxy_host: str = None, proxy_port: int = None, size: int = None,
    progress: Progress = None, headers: dict = None, callback: Callable = None
) -> int:
    """
    Download the file from the url.

    :param url: File url
    :param save_path: Save path
    :param filename: Save filename.
    :param proxy_host: Proxy server port
    :param proxy_port: Proxy server address
    :param size: File size in kb. Some websites may not return the content size, in which case you should give the size to make progress show correctly.
    :param progress: A Progress object to display download progress
    :param headers: Self-defined http headers.
    :param callback: Callable object that will be called every iteration of `res.iter_content`. `callback` should accept two params: `total_size` and `step_size`.
                     `callback` can return a non-zero and non-None value to stop download.
    :return: res status code
    """
    # check the save path
    if not exists(save_path):
        makedirs(save_path)

    if save_path[-1] != '/':
        save_path += '/'

    # generate proxy setting
    if proxy_host is None or proxy_port is None:
        proxy_setting = None
    else:
        proxy_setting = {
            "http": f"{proxy_host}:{proxy_port}",
            "https": f"{proxy_host}:{proxy_port}"
        }

    # check headers
    if not isinstance(headers, dict):
        headers = HEADERS

    # use debug to log url because usually user only cares if download successfully.
    logger.debug(f"Downloading file from: {url}")

    # loop 5 times
    step = 0
    res = False
    task_id = None
    while step < 4:
        try:
            res = get(url, headers=headers, allow_redirects=True, proxies=proxy_setting, stream=True)
            # check code, if it isn't 200, retry
            if res.status_code == 404:
                return 404
            elif res.status_code != 200:
                sleep(1)
                step += 1
            else:
                # set progress bar
                if size is None:
                    size = res.headers.get('content-length')
                if size is None:
                    size = 1000
                if progress is not None:
                    task_id = progress.add_task(f"[red]{filename}[red]", total=int(size))
                # save data to a temp file in case download is terminated
                temp_filename = f"{filename}.part"
                with open(save_path + temp_filename, 'wb') as f:
                    for data in res.iter_content(chunk_size=4096):
                        f.write(data)
                        if progress is not None:
                            progress.update(task_id, advance=len(data))
                        if callback is not None:
                            return_value = callback(int(size), len(data))
                            # check return code of callback
                            if return_value is not None and return_value != 0:
                                logger.info(f"Stop downloading by callback")
                                return -1
                # rename
                if exists(save_path + filename):
                    # remove the old file
                    remove(save_path + filename)
                rename(save_path + temp_filename, save_path + filename)
                break
        except (ConnectTimeout, ProxyError, SSLError) as error:
            logger.error(f"Error \"{error}\" occurred, retry...")
            sleep(1)
            step += 1
            continue
        except RequestException as error:
            logger.error(f"Unexpected error occurred: {error}, stop download")
            # other exception
            return -1

    # check if download successfully
    if isinstance(res, bool):
        # raise Exception(f"Fail to download data after retrying 5 times from {url}")
        return -1
    # if the user gives the progress, maybe we are in a loop. so remove finished progress tasks
    if progress is not None and task_id is not None:
        progress.stop_task(task_id)
        progress.remove_task(task_id)

    return res.status_code


def download_ftp(
    ftp_url: str, save_path: str, filename: str, user: str = None, passwd: str = None, proxy_host: str = None, proxy_port: int = None, progress_num=1,
    show_progress=True, progress: Progress = None, callback: Callable = None, timeout=720, connection_timeout=30, force_download=False
) -> bool:
    """
    download file from ftp server.

    :param ftp_url: file ftp url
    :param save_path: data save path
    :param filename: filename
    :param user: ftp server username
    :param passwd: ftp server password
    :param proxy_host: proxy setting.
    :param proxy_port: proxy setting.
    :param progress_num: The number of progresses that is used to download the file.
    :param show_progress: If True, display download progress in console
    :param progress: Progress object to display download progress
    :param callback: Callable object that will be called every iteration of `res.iter_content`. `callback` should accept two params: `total_size` and `step_size`.
                     `callback` can return a non-zero and non-None value to stop download.
    :param timeout: The max time which could be used to download file. Download will be terminated once the max time reached. Set it to -1 for no limitation. Unit: seconds.
    :param connection_timeout:
    :param force_download: If ignore the existing data and re-download it.
    :type force_download: bool
    :return: bool value, True if download successfully
    """
    if show_progress:
        if isinstance(progress, Progress):
            return _download_ftp(
                ftp_url, save_path, filename, user=user, passwd=passwd, proxy_host=proxy_host, proxy_port=proxy_port,
                progress=progress, callback=callback, timeout=timeout, connection_timeout=connection_timeout, progress_num=progress_num,
                force_download=force_download
            )
        else:
            with PROGRESS:
                # remove finished task
                for task in PROGRESS.task_ids:
                    PROGRESS.remove_task(task)
                return _download_ftp(
                    ftp_url, save_path, filename, user=user, passwd=passwd, proxy_host=proxy_host, proxy_port=proxy_port,
                    progress=PROGRESS, callback=callback, timeout=timeout, connection_timeout=connection_timeout, progress_num=progress_num,
                    force_download=force_download
                )
    else:
        return _download_ftp(
            ftp_url, save_path, filename, user=user, passwd=passwd, proxy_host=proxy_host, proxy_port=proxy_port, callback=callback,
            timeout=timeout, connection_timeout=connection_timeout, progress_num=progress_num,
            force_download=force_download
        )


def _download_ftp(
    ftp_url: str, save_path: str, filename: str, user: str = None, passwd: str = None, proxy_host: str = None, proxy_port: int = None, progress_num=1,
    progress: Progress = None, callback: Callable = None, timeout: int = 720, connection_timeout=30, force_download=False
) -> bool:
    """
    Download file from ftp server.

    :param ftp_url: File url
    :param save_path: Data save path
    :param filename: Filename
    :param user: FTP server username.
    :param passwd: FTP server password.
    :param proxy_host: Proxy host. Only support socks5.
    :param proxy_port: Proxy port.
    :param progress_num: The number of progresses that is used to download the file.
    :param progress: Progress object to display download progress.
    :param callback: Callable object that will be called every iteration of `res.iter_content`. `callback` should accept two params: `total_size` and `step_size`.
                     `callback` can return a non-zero and non-None value to stop download.
    :param timeout: The max time which could be used to download file. Download will be terminated once the max time reached. Set it to -1 for no limitation. Unit: seconds.
    :param connection_timeout:
    :param force_download: If ignore the existing data and re-download it.
    :type force_download: bool
    :return:  value, True if download successfully
    """
    # If username and password are given, we need to generate new url
    if user is not None and passwd is not None:
        url = ftp_url.split("ftp://")[1]
        ftp_url = f"ftp://{user}:{passwd}@{url}"

    downloader = CurlDownloader(ftp_url, save_path, filename, proxy_host, proxy_port, timeout, connection_timeout)
    return downloader.start(progress=progress, callback=callback, process_num=progress_num, force_download=force_download)


def mask_land(data: xr.DataArray, landmask: Union[str, xr.DataArray, None] = None) -> xr.DataArray:
    """
    Mask the land value in input data.

    :param data: DataArray data
    :param landmask: NetCDF landmask data file or a DataArray object.
           If None, use Python package `global-land-mask` to generate land mask.
    :return: Data with land masked
    """
    # check data, we require data contains dimensions called "latitude" and "longitude"
    if "latitude" not in data.dims or "longitude" not in data.dims:
        logger.error(f"Can't found dimension `latitude` and `longitude` in data.")
        logger.error(f"If your data contains coordinates, considering rename it to `latitude` and `longitude`")
        raise KeyError(f"Can't found dimension `latitude` and `longitude` in data.")

    if isinstance(landmask, str):
        assert exists(landmask), f"landmask file {landmask} does not exist"

        # read landmask file
        landmask = xr.open_dataset(landmask)
        landmask = landmask['landmask']

    if isinstance(landmask, xr.DataArray):
        # check the length of each dimension in both files
        assert data['latitude'].size == landmask['latitude'].size, "latitude dimension length does not match"
        assert data['longitude'].size == landmask['longitude'].size, "longitude dimension length does not match"
        landmask = landmask.to_numpy()

    else:
        # generate land mask use `global-land-mask`
        latitude = data["latitude"]
        longitude = data["longitude"]
        longitude, latitude = np.meshgrid(longitude, latitude)
        landmask = is_land(latitude, longitude)

    # mask the points in data
    np_data = data.to_numpy()
    np_data[landmask] = np.nan

    data = xr.DataArray(name=data.name, data=np_data, dims=data.dims, coords=data.coords, attrs=data.attrs)
    return data


__all__ = ['solar_altitude_zenith', 'atmosphere_visibility', 'seafog_top_height', 'solar_altitude_zenith_ephem', 'CurlDownloader',
           'solar_altitude_zenith_formula', 'decompress_file', 'download_url', 'download_ftp', 'logger', 'PROGRESS', 'mask_land']



"""
Created on Thu Jun  1 16:20 2023
This Python file is used to read data from pgm picture and calculate the corresponding brightness temperature with the loop up table file
"""

from datetime import datetime
from os import makedirs
from os.path import exists
from time import sleep
from typing import Tuple, Callable, Iterable, Union

import numpy as np
from pgm_reader import Reader
from rich.progress import Progress
from xarray import DataArray

from .res import *
from .res import get_default_table_file
from .utils import decompress_file, download_url, logger

ROOT_URL = "http://weather.is.kochi-u.ac.jp/sat"
# GAME/{year}/{month}/{data-type}
PGM_FILE_URL = "GAME/{}/{}/{}"
# {header}{time-string}{data-type}.pgm.gz, i.e. MT1R10122010IR1.pgm.gz
PGM_FILE_NAME = "{}{}{}.pgm.gz"
# CAL/{year}/{month}
TABLE_FILE_URL = "CAL/{}/{}"
# {header}{time-string}CAL.dat.gz, i.e. MTS211050106CAL.dat.gz
TABLE_FILE_NAME = "{}{}CAL.dat.gz"
# from now on the header change from MT1R to MTS2
HEADER_CHANGE_1 = datetime(2010, 12, 22, 3, 0)
# from now on the header change from MTS2 to HMW8
HEADER_CHANGE_2 = datetime(2015, 7, 1, 2)

# month dict
MONTH = {'01': 'Jan', '02': 'Feb', '03': 'Mar', '04': 'Apr', '05': 'May', '06': 'Jun', '07': 'Jul', '08': 'Aug', '09': 'Sep', '10': 'Oct', '11': 'Nov', '12': 'Dec'}


def generate_data_info(date: datetime, header: str, data_type: str, save_path: str) -> Tuple[str, str, str]:
    """
    generate data filename, download url and save path.

    :param date: for example, 2022-05-19 12:00, UTC time
    :param header: filename header
    :param data_type: valid value: IR1, IR2, IR3, IR4, VIS
    :param save_path: data save path
    :return: tuple(filename, download url, save path)
    """
    pre_url = PGM_FILE_URL.format(date.strftime("%Y"), MONTH[date.strftime("%m")], data_type)
    filename = PGM_FILE_NAME.format(header, date.strftime("%Y%m%d%H")[2:], data_type)
    url = f"{ROOT_URL}/{pre_url}/{filename}"
    save_path = f"{save_path}/{filename}"

    return filename, url, save_path


def generate_table_info(date: datetime, header: str, save_path: str) -> Tuple[str, str, str]:
    """
    generate table filename, download url and save path.

    :param date: for example, 2022-05-19 12:00, UTC time
    :param header: filename header
    :param save_path: data save path
    :return:tuple(filename, download url, save path)
    """
    pre_url = TABLE_FILE_URL.format(date.strftime("%Y"), MONTH[date.strftime("%m")])
    filename = TABLE_FILE_NAME.format(header, date.strftime("%Y%m%d%H")[2:])
    url = f"{ROOT_URL}/{pre_url}/{filename}"
    save_path = f"{save_path}/{filename}"

    return filename, url, save_path


def get_file_header(date: datetime) -> str:
    """
    get the data file header based on data date.

    :param date: data date
    :return: data filename header
    """
    if date < HEADER_CHANGE_1:
        return "MT1R"
    elif HEADER_CHANGE_1 <= date < HEADER_CHANGE_2:
        return "MTS2"
    else:
        return "HMW8"


def pgm_parser(pgm_filename: str, table_filename: str, data_type='IR1') -> DataArray:
    """
    Parse the pgm file and calculate the corresponding brightness temperature with the loop up table file.
    For VIS table calculation, VIS data is calibrated with 0:=0% and 255:=100% (albedo) and can be linearly interpolated.
    See http://weather.is.kochi-u.ac.jp/archive-e.html

    :param pgm_filename: PGM filename
    :param table_filename: table filename
    :param data_type: 'IR1', 'IR4', 'VIS' or 'VIS-b'.
    Note that if the data type is 'VIS', you need to give specific VIS temperature loop up table, or you can use 'VIS-b' instead to use built-in table.
    :return: PGM data.
    """
    assert exists(pgm_filename), f"File Not Exists: {pgm_filename}"
    if data_type != "VIS-b":
        assert exists(table_filename), f"File Not Exists: {table_filename}"

    # if the data type is VIS-b, use built-in table file
    if data_type == 'VIS-b':
        logger.debug(f'Use built in VIS look up table which is in: {VIS_table_file}, you can change it as you like')
        table_filename = VIS_table_file

    # check if table file needs to process
    with open(table_filename, 'r') as f:
        table_data = f.readlines()
    # try to extract the table value
    try:
        table_value = {i: float(table_data[i][:-1]) for i in range(len(table_data))}
    except ValueError:
        # raw table file, need to get the corresponding value
        logger.debug('Raw table file, need to be processed')
        logger.debug(f"Your data type is {data_type}")
        # check if the data type is VIS
        if 'VIS' not in data_type:
            # get the corresponding data type line
            table_data = [x for x in table_data if data_type in x]
            # remove the first useless line
            table_data = table_data[1:]
            # split and get the latest string as the value
            table_data = [x.split()[-1] for x in table_data]
            table_value = {i: float(table_data[i]) for i in range(len(table_data))}
        else:
            logger.debug(f'Use built in VIS look up table which is in: {VIS_table_file}, you can change it as you like')
            # read pre-calculated value
            with open(VIS_table_file, 'r') as f:
                table_data = f.readlines()
                table_value = {i: float(table_data[i][:-1]) for i in range(len(table_data))}

    # read pgm file
    reader = Reader()
    data = reader.read_pgm(pgm_filename)
    new_data = data.copy().astype(float)
    # get the corresponding temperature
    keys = list(table_value.keys())
    for key in keys:
        new_data[data == key] = table_value[key]

    # turn data upside down
    new_data = new_data[::-1, :]

    # generate longitude and latitude
    longitude = np.linspace(70.15, 160.15, 1800)
    latitude = np.linspace(-20.15, 69.85, 1800)
    # generate xarray DataArray object
    new_data = DataArray(name=data_type, data=new_data, dims=['latitude', 'longitude'], coords={
        'latitude': latitude,
        'longitude': longitude
    }, attrs={
        'units': 'K'
    })

    return new_data


def pgm_find_data(date: str, save_path: str, data_type: Union[str, Iterable[str]] = 'default', built_in=True, proxy_host: str = None, proxy_port: int = None,
                  progress: Progress = None, headers: dict = None, show_progress=True, callback: Callable = None) -> Tuple[str, ...]:
    """
    Use this function to find data in the specify data path.
    If data is not find, this will download it automatically.

    :param date: for example, 2022-05-19 12:00, UTC time
    :param save_path: data save path
    :param data_type: valid value: IR1, IR2, IR3, IR4, VIS, all, default (IR1, IR4 and VIS), or a list contains multiple data types.
    :param built_in: if table file doesn't exist in server, use built-in table file
    :param proxy_host: http and https proxy server address
    :param proxy_port: http and https proxy server port
    :param progress: Progress object to display download progress
    :param headers: Self-defined http headers.
    :param show_progress: If True, display download progress in console.
    :param callback: Callable object that will be called every iteration of `res.iter_content`.
    `callback` should accept two params: `total_size` and `step_size`
    :return: data paths, including the data of your input type and their table file

    Download IR1, IR4 and VIS data:

    >>> pgm_find_data("2024-07-29 00:00", "data", show_progress=False)
    ('data/HMW824072900IR1.pgm', 'data/HMW824072900IR4.pgm', 'data/HMW824072900VIS.pgm', 'data/HMW824072900CAL.dat')

    Download IR 1-4 data:

    >>> pgm_find_data("2024-07-29 00:00", "data", data_type=["IR1", "IR2", "IR3", "IR4"], show_progress=False)
    ('data/HMW824072900IR1.pgm', 'data/HMW824072900IR2.pgm', 'data/HMW824072900IR3.pgm', 'data/HMW824072900IR4.pgm', 'data/HMW824072900CAL.dat')

    """
    # check data path, if it doesn't exist, download data directly
    if not exists(save_path):
        makedirs(save_path)

    # get data type and corresponding data header
    date = datetime.strptime(date, "%Y-%m-%d %H:%M")
    if data_type == 'default':
        data_type_list = ['IR1', 'IR4', 'VIS']
    elif data_type == 'all':
        data_type_list = ['IR1', 'IR2', 'IR3', 'IR4', 'VIS']
    elif isinstance(data_type, str):
        data_type_list = [data_type]
    else:
        data_type_list = data_type
    header = get_file_header(date)

    # store data path
    data_path_list = []
    # store missing data name and url
    miss_data_info = {}

    # find exist data
    for data_type in data_type_list:
        filename, file_url, file_path = generate_data_info(date, header, data_type, save_path)
        if not exists(file_path[:-3]):
            miss_data_info[filename] = file_url
        else:
            data_path_list.append(file_path[:-3])

    # download missing data
    if len(miss_data_info) > 0:
        for filename in miss_data_info:
            # download file and decompress it
            # logger.debug(f"Downloading file from {miss_data_info[filename]}")
            code = download_url(miss_data_info[filename], save_path, filename, proxy_host=proxy_host, proxy_port=proxy_port, headers=headers,
                                show_progress=show_progress, progress=progress, callback=callback)
            if code == 404:
                logger.warning(f"The file {filename} doesn't exist in server (status 404), following download tasks will be terminated")
                # generate empty results
                return tuple(["" for _ in data_type_list] + [""])
            elif code != 200:
                logger.error(f"Failed to download file {filename}, status code is {code}. "
                             f"May be you can try again later, or check if this url is right: {miss_data_info[filename]}")
                raise ConnectionError

            decompress_file(f"{save_path}/{filename}")

            data_path_list.append(f"{save_path}/{filename[:-3]}")

            # sleep one second for humanitarianism
            sleep(0.5)

    data_path_list.sort()

    # find exist table file
    filename, table_url, table_path = generate_table_info(date, header, save_path)
    # download table file
    if not exists(table_path[:-3]):
        # logger.debug(f"Downloading file from {table_url}")
        code = download_url(table_url, save_path, filename, proxy_host=proxy_host, proxy_port=proxy_port, headers=headers,
                            show_progress=show_progress, progress=progress, callback=callback)
        if code == 404 or code == -1:
            if built_in:
                logger.warning(f"The file {filename} doesn't exist in server (status {code}), use built-in table file by default. "
                               f"You can pass `built_in=False` to disable this feature")
                table_path = get_default_table_file(header)
                data_path_list.append(table_path)
        elif code != 200:
            logger.error(f"Failed to download file {filename}, status code is {code}. May be you can try again later, or check if this url is right: {table_url}")
            raise ConnectionError
        else:
            decompress_file(table_path)
            data_path_list.append(table_path[:-3])
    else:
        data_path_list.append(table_path[:-3])

    return tuple(data_path_list)


__all__ = ['pgm_parser', 'pgm_find_data', 'generate_data_info', 'generate_table_info', 'get_file_header']



"""
This module contains all paths of resource files and global variables.
"""

from os.path import abspath, dirname, exists
from sys import platform

res_path = abspath(dirname(__file__))
VIS_table_file = f"{res_path}/vis_table.txt"
LANDMASK_file = f"{res_path}/landmask.nc"
HMW8_file = f"{res_path}/HMW8.txt"
MTS2_file = f"{res_path}/MTS2.txt"

# save file in /tmp on linux
if platform == "linux":
    AVHRR_CACHE_FOLDER = f"/tmp/seafog/AVHRR/"
    TIFF_CACHE = f"/tmp/seafog/cache.tiff"
else:
    AVHRR_CACHE_FOLDER = f"{res_path}/AVHRR/"
    TIFF_CACHE = f"{res_path}/cache.tiff"


def get_default_table_file(header: str) -> str:
    """
    get table file based on the header.

    :param header: valid header: "MTS2" or "HMW8".
    :return: built-in table file path.
    """
    filepath = f"{res_path}/{header}.txt"
    if exists(filepath):
        return filepath
    else:
        return ""


__all__ = ['res_path', 'VIS_table_file', 'LANDMASK_file', 'HMW8_file', 'MTS2_file', 'get_default_table_file', 'AVHRR_CACHE_FOLDER', 'TIFF_CACHE']
